"""Stream type classes for tap-clickup."""
from pathlib import Path
from typing import Optional, Any, Dict, cast, Iterable
import datetime
import pendulum
import requests
from singer_sdk.helpers.jsonpath import extract_jsonpath
from tap_clickup.client import ClickUpStream

SCHEMAS_DIR = Path(__file__).parent / Path("./schemas")


class TeamsStream(ClickUpStream):
    """Teams"""

    name = "team"
    path = "/team"
    primary_keys = ["id"]
    replication_key = None
    # Optionally, you may also use `schema_filepath` in place of `schema`:
    schema_filepath = SCHEMAS_DIR / "team.json"
    records_jsonpath = "$.teams[*]"

    def get_child_context(self, record: dict, context: Optional[dict]) -> dict:
        """Return a context dictionary for child streams."""
        return {
            "team_id": record["id"],
        }


class SpacesStream(ClickUpStream):
    """Spaces"""

    name = "space"
    path = "/team/{team_id}/space"
    primary_keys = ["id"]
    replication_key = None
    # Optionally, you may also use `schema_filepath` in place of `schema`:
    schema_filepath = SCHEMAS_DIR / "space.json"
    records_jsonpath = "$.spaces[*]"
    parent_stream_type = TeamsStream
    partitions = [{"archived":"true"},{"archived":"false"}]


    def get_child_context(self, record: dict, context: Optional[dict]) -> dict:
        """Return a context dictionary for child streams."""
        return {
            "space_id": record["id"],
        }

class FoldersStream(ClickUpStream):
    """Folders"""

    name = "folder"
    path = "/space/{space_id}/folder"
    primary_keys = ["id"]
    replication_key = None
    schema_filepath = SCHEMAS_DIR / "folder.json"
    records_jsonpath = "$.folders[*]"
    parent_stream_type = SpacesStream
    partitions = [{"archived":"true"},{"archived":"false"}]

    def get_child_context(self, record: dict, context: Optional[dict]) -> dict:
        """Return a context dictionary for child streams."""
        return {
            "folder_id": record["id"],
        }


class FolderListsStream(ClickUpStream):
    """Lists"""

    name = "folder_list"
    path = "/folder/{folder_id}/list"
    primary_keys = ["id"]
    replication_key = None
    schema_filepath = SCHEMAS_DIR / "list.json"
    records_jsonpath = "$.lists[*]"
    parent_stream_type = FoldersStream
    partitions = [{"archived":"true"},{"archived":"false"}]

    def get_child_context(self, record: dict, context: Optional[dict]) -> dict:
        """Return a context dictionary for child streams."""
        return {
            "list_id": record["id"],
        }


class FolderlessListsStream(ClickUpStream):
    """Folderless Lists"""

    name = "folderless_list"
    path = "/space/{space_id}/list"
    primary_keys = ["id"]
    replication_key = None
    schema_filepath = SCHEMAS_DIR / "list.json"
    records_jsonpath = "$.lists[*]"
    parent_stream_type = SpacesStream
    partitions = [{"archived":"true"},{"archived":"false"}]

    def get_child_context(self, record: dict, context: Optional[dict]) -> dict:
        """Return a context dictionary for child streams."""
        return {
            "list_id": record["id"],
        }


class TaskTemplatesStream(ClickUpStream):
    """TaskTemplates"""

    name = "task_template"
    path = "/team/{team_id}/task_template?page=0"
    primary_keys = ["id"]
    replication_key = None
    schema_filepath = SCHEMAS_DIR / "task_template.json"
    records_jsonpath = "$.templates[*]"
    parent_stream_type = TeamsStream


class GoalsStream(ClickUpStream):
    """Goals"""

    name = "goal"
    path = "/team/{team_id}/goal"
    primary_keys = ["id"]
    replication_key = None
    schema_filepath = SCHEMAS_DIR / "goal.json"
    records_jsonpath = "$.goals[*]"
    parent_stream_type = TeamsStream


class TagsStream(ClickUpStream):
    """Tags"""

    name = "tag"
    path = "/space/{space_id}/tag"
    primary_keys = ["name"]
    replication_key = None
    schema_filepath = SCHEMAS_DIR / "tag.json"
    records_jsonpath = "$.tags[*]"
    parent_stream_type = SpacesStream


class SharedHierarchyStream(ClickUpStream):
    """SharedHierarchy"""

    name = "shared_hierarchy"
    path = "/team/{team_id}/shared"
    primary_keys = []
    replication_key = None
    schema_filepath = SCHEMAS_DIR / "shared.json"
    records_jsonpath = "$.shared"
    parent_stream_type = TeamsStream


class FolderlessCustomFieldsStream(ClickUpStream):
    """CustomField from folderless lists"""

    name = "folderless_customfield"
    path = "/list/{list_id}/field"
    primary_keys = ["id"]
    replication_key = None
    schema_filepath = SCHEMAS_DIR / "custom_field.json"
    records_jsonpath = "$.fields[*]"
    parent_stream_type = FolderlessListsStream


class FolderCustomFieldsStream(ClickUpStream):
    """CustomFields from foldered lists"""

    name = "folder_customfield"
    path = "/list/{list_id}/field"
    primary_keys = ["id"]
    replication_key = None
    schema_filepath = SCHEMAS_DIR / "custom_field.json"
    records_jsonpath = "$.fields[*]"
    parent_stream_type = FolderListsStream

class TasksStream(ClickUpStream):
    """Tasks Stream"""

    name = "task"
    #Date_updated_gt is greater than or equal to not just greater than
    path = "/team/{team_id}/task?include_closed=true&subtasks=true"
    primary_keys = ["id"]
    #replication_key = "date_updated" 
    #is_sorted = True
    #ignore_parent_replication_key = True
    schema_filepath = SCHEMAS_DIR / "task.json"
    records_jsonpath = "$.tasks[*]"
    parent_stream_type = TeamsStream
    partitions = [{"archived":"true"},{"archived":"false"}]
   
    initial_replication_key_dict = {}

    def initial_replication_key(self, context) -> int:
        path = self.get_url(context) + context.get("archived")
        key_cache: Optional[int] = self.initial_replication_key_dict.get(path, None)
        if key_cache is None:
            key_cache = self.get_starting_replication_key_value(context)
            self.initial_replication_key_dict[path] = key_cache
        assert key_cache is not None
        return key_cache
    
    def get_url_params(
        self, context: Optional[dict], next_page_token: Optional[Any]
    ) -> Dict[str, Any]:
        """Return a dictionary of values to be used in URL parameterization."""
        params: dict = {}
        if next_page_token:
            params["page"] = next_page_token
        params["archived"] = context.get("archived")
        params["order_by"] = "updated"
        params["reverse"] = "true"
        params["date_updated_gt"] = 0
        return params

    def get_starting_replication_key_value(
        self, context: Optional[dict]
    ) -> Optional[int]:
        """Return starting replication key value. """
        if self.replication_key:
            state = self.get_context_state(context)
            replication_key_value = state.get("replication_key_value")
            if replication_key_value and self.replication_key == state.get(
                "replication_key"
            ):
                return replication_key_value
            if "start_date" in self.config:
                datetime_startdate = cast(
                    datetime.datetime, pendulum.parse(self.config["start_date"])
                )
                startdate_seconds_after_epoch = int(
                    datetime_startdate.replace(tzinfo=datetime.timezone.utc).timestamp()
                )
                return startdate_seconds_after_epoch
            else:
                self.logger.info(
                    """Setting replication value to 0 as there wasn't a
                    start_date provided in the config."""
                )
                return 0
        return None

    def get_next_page_token(
        self, response: requests.Response, previous_token: Optional[Any]
    ) -> Optional[Any]:
        """Return the page number, Null if we should stop going to the next page."""
        newtoken = None
        recordcount = 0
        if previous_token is None:
            previous_token = 0

        for _ in extract_jsonpath(self.records_jsonpath, input=response.json()):
            recordcount = recordcount + 1

        # I wonder if a better approach is to just check for 0 records and stop
        # For now I'll follow the docs verbatium
        # From the api docs, https://clickup.com/api.
        # you should check list limit against the length of each response
        # to determine if you are on the last page.
        if recordcount == 100:
            newtoken = previous_token + 1
        else:
            newtoken = None

        return newtoken



"""ClickUp tap class."""

from typing import List

from singer_sdk import Tap, Stream
from singer_sdk import typing as th

from tap_clickup.streams import (
    TeamsStream,
    SpacesStream,
    FoldersStream,
    FolderListsStream,
    FolderlessListsStream,
    TaskTemplatesStream,
    GoalsStream,
    TagsStream,
    SharedHierarchyStream,
    TasksStream, 
    FolderCustomFieldsStream,
    FolderlessCustomFieldsStream,
)

STREAM_TYPES = [
    TeamsStream,
    SpacesStream,
    FoldersStream,
    FolderListsStream,
    FolderlessListsStream,
    TaskTemplatesStream,
    GoalsStream,
    TagsStream,
    SharedHierarchyStream,
    TasksStream,
    FolderCustomFieldsStream,
    FolderlessCustomFieldsStream,
]


class TapClickUp(Tap):
    """ClickUp tap class."""

    name = "tap-clickup"

    config_jsonschema = th.PropertiesList(
        th.Property("api_token", th.StringType, required=True),
    ).to_dict()

    def discover_streams(self) -> List[Stream]:
        """Return a list of discovered streams."""
        return [stream_class(tap=self) for stream_class in STREAM_TYPES]



"""REST client handling, including ClickUpStream base class."""

from typing import Any, Optional, Iterable, cast, Dict
from pathlib import Path
from datetime import datetime
import time
import requests
import backoff
from requests.exceptions import RequestException
from singer_sdk.helpers.jsonpath import extract_jsonpath
from singer_sdk.streams import RESTStream

SCHEMAS_DIR = Path(__file__).parent / Path("./schemas")


class ClickUpStream(RESTStream):
    """ClickUp stream class."""

    url_base = "https://api.clickup.com/api/v2"
    records_jsonpath = "$[*]"  # Or override `parse_response`.
    next_page_token_jsonpath = "$.next_page"  # Or override `get_next_page_token`.
    
    def get_url_params(
        self, context: Optional[dict], next_page_token: Optional[Any]
    ) -> Dict[str, Any]:
        """Return a dictionary of values to be used in URL parameterization."""
        params: dict = {}
        if next_page_token:
            params["page"] = next_page_token
        if context:
            params["archived"] = context.get("archived")
        return params

    @property
    def http_headers(self) -> dict:
        """Return the http headers needed."""
        headers = {}
        if "user_agent" in self.config:
            headers["User-Agent"] = self.config.get("user_agent")

        headers["Authorization"] = self.config.get("api_token")
        return headers

    @backoff.on_exception(
        backoff.expo,
        (requests.exceptions.RequestException),
        max_tries=5,
        giveup=lambda e: e.response is not None
        and (e.response.status_code != 429 and 400 <= e.response.status_code < 500),
        factor=2,
    )
    def _request_with_backoff(
        self, prepared_request, context: Optional[dict]
    ) -> requests.Response:
        response = self.requests_session.send(prepared_request)
        if self._LOG_REQUEST_METRICS:
            extra_tags = {}
            if self._LOG_REQUEST_METRIC_URLS:
                extra_tags["url"] = cast(str, prepared_request.path_url)
            self._write_request_duration_log(
                # Shows useful incremental debugging info for clickup
                # There is no sensitive data here
                endpoint=prepared_request.path_url,
                response=response,
                context=context,
                extra_tags=extra_tags,
            )
        if response.status_code in [401, 403]:
            self.logger.info("Failed request for {}".format(prepared_request.url))
            self.logger.info(
                f"Reason: {response.status_code} - {str(response.content)}"
            )
            raise RuntimeError(
                "Requested resource was unauthorized, forbidden, or not found."
            )
        elif response.status_code == 429:
            # ClickOff api limits to 100 requests/min.
            # There's probably a nicer way to use this with the backoff library
            # Does not rely on any local time information

            reset_epoch: int = int(response.headers.get("X-RateLimit-Reset"))
            date = response.headers.get("Date")
            dformat = "%a, %d %b %Y %H:%M:%S %Z"
            epoch = datetime(1970, 1, 1)
            currentEpoch = (datetime.strptime(date, dformat) - epoch).total_seconds()
            waitTime = reset_epoch - currentEpoch
            self.logger.info(
                f"API Limit reached, waiting {waitTime} seconds and will try again."
            )
            if waitTime > 120:
                self.logger.warning(
                    "Wait time is more than 2 minutes, Waiting 60s and trying again."
                )
                time.sleep(60)
            else:
                time.sleep(waitTime)

            raise RequestException

        elif response.status_code >= 400:
            raise RuntimeError(
                f"Error making request to API: {prepared_request.url} "
                f"[{response.status_code} - {str(response.content)}]".replace(
                    "\\n", "\n"
                )
            )
        self.logger.debug("Response received successfully.")
        return response

    def parse_response(self, response: requests.Response) -> Iterable[dict]:
        """Parse the response and return an iterator of result rows."""
        yield from extract_jsonpath(self.records_jsonpath, input=response.json())

    def from_parent_context(self, context: dict):
        """Default is to return the dict passed in"""
        if(self.partitions is None): return context
        else: 
            #Was going to copy the partitions, but the _sync call, forces us
            #To use partitions, instead of being able to provide a list of contexts
            #Ideally we wouldn't mutate partitions here, and we'd just provide
            #A copy of partitions with context merged so we don't have side effects
            for partition in self.partitions:
                partition.update(context.copy()) #Add copy of context to partition
            return None #Context now handled at the partition level
    
    def _sync_children(self, child_context: dict) -> None:
        for child_stream in self.child_streams:
            if child_stream.selected or child_stream.has_selected_descendents:
                child_stream.sync(child_stream.from_parent_context(context=child_context))




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tap_clickup', 'tap_clickup.tests']

package_data = \
{'': ['*'], 'tap_clickup': ['schemas/*']}

install_requires = \
['requests>=2.25.1,<3.0.0', 'singer-sdk==0.3.13']

entry_points = \
{'console_scripts': ['tap-clickup = tap_clickup.tap:TapClickUp.cli']}

setup_kwargs = {
    'name': 'tap-clickup',
    'version': '0.0.14',
    'description': 'tap-clickup is a Singer tap for ClickUp by AutoIDM',
    'long_description': "# `tap-clickup` ![Build and Tests](https://github.com/AutoIDM/tap-clickup/actions/workflows/ci.yml/badge.svg?branch=main) [![PyPI download month](https://img.shields.io/pypi/dm/tap-clickup.svg)](https://pypi.python.org/pypi/tap-clickup/) \n`tap-clickup` is a Singer tap for ClickUp.\n\n## Capabilities\n\n* `sync`\n* `catalog`\n* `state`\n* `discover`\n\n## Settings\n\n| Setting   | Required | Default | Description |\n|:----------|:--------:|:-------:|:------------|\n| api_token | True     | None    | Example: 'pk_12345'            |\n\nA full list of supported settings and capabilities is available by running: `tap-clickup --about`\n\n### Getting an API Token\n\n1. Login at https://app.clickup.com/\n1. Click your icon at the bottom left of the screen\n1. Click My Settings\n1. Click Apps (Bottom leftish of screen under the My Apps sub section)\n1. At the top of the screen is an API Token. This can be used in the api_token location. \n    * This is a personal token, it's fine to use a personal token as this tap is only for the business that's using the data. \n\n## Clickup Replication\nStart Date is used for the initial updated at value for the updated at field with tasks. \n\nLet's say that you only want tasks that have been updated in the last year. To accomplish this you would pass in a start date of the first of this year!\n\n## Clickup Table Schemas\nNote that the most up to date information is located in tap_clickup/streams.py. We will try to keep these docs updated\n\n### Teams\n- Table name: team\n- Description: Teams Data, each user can be in multiple teams\n- Primary key column(s):  id\n- Replicated fully or incrementally: No\n- Bookmark column(s): N/A\n- Link to API endpoint documentation: [Teams](https://jsapi.apiary.io/apis/clickup20/reference/0/teams/get-teams.html)\n\n### Spaces\n- Table name: space\n- Description: Each team has multiple spaces\n- Primary key column(s):  id\n- Replicated fully or incrementally: No\n- Bookmark column(s): N/A\n- Link to API endpoint documentation: [Spaces](https://jsapi.apiary.io/apis/clickup20/reference/0/spaces.html)\n\n### Folders\n- Table name: folder\n- Description: Each space can have multiple folders\n- Primary key column(s):  id\n- Replicated fully or incrementally: No\n- Bookmark column(s): N/A\n- Link to API endpoint documentation: [Folders](https://jsapi.apiary.io/apis/clickup20/reference/0/folders.html)\n\n### Folder Lists\n- Table name: folder_list\n- Description: Each Folder can have multiple lists\n- Primary key column(s):  id\n- Replicated fully or incrementally: No\n- Bookmark column(s): N/A\n- Link to API endpoint documentation:[Folder Lists](https://jsapi.apiary.io/apis/clickup20/reference/0/lists/get-lists.html)\n\n\n### Folderless Lists\n- Table name: folderless_list\n- Description: Some lists do not exist in a folder\n- Primary key column(s):  id\n- Replicated fully or incrementally: No\n- Bookmark column(s): N/A\n- Link to API endpoint documentation: [Folderless Lists](https://jsapi.apiary.io/apis/clickup20/reference/0/lists/get-folderless-lists.html)\n\n### Task Templates\n- Table name: task_template\n- Description: Tasks can be templated for any reason you can imagine!\n- Primary key column(s):  id\n- Replicated fully or incrementally: No\n- Bookmark column(s): N/A\n- Link to API endpoint documentation: [Task Templates](https://jsapi.apiary.io/apis/clickup20/reference/0/task-templates/get-task-templates.html)\n\n### Goals\n- Table name: goal \n- Description: Teams can set goals for themselves\n- Primary key column(s):  id\n- Replicated fully or incrementally: No\n- Bookmark column(s): N/A\n- Link to API endpoint documentation: [Goals](https://jsapi.apiary.io/apis/clickup20/reference/0/goals/get-goals.html)\n\n### Tags\n- Table name: tag\n- Description: Each space can have multiple tags\n- Primary key column(s):  id\n- Replicated fully or incrementally: No\n- Bookmark column(s): N/A\n- Link to API endpoint documentation: [Tags](https://jsapi.apiary.io/apis/clickup20/reference/0/tags/get-space-tags.html)\n\n### Shared Hierarchy\n- Table name: shared_hierarchy\n- Description: Returns all resources you have access to where you don't have access to its parent. For example, if you have a access to a shared task, but don't have access to its parent list, it will come back in this request.\n- Primary key column(s): (No primary key column)\n- Replicated fully or incrementally: No\n- Bookmark column(s): N/A\n- Link to API endpoint documentation: [Shared Hierarchy](https://jsapi.apiary.io/apis/clickup20/reference/0/shared-hierarchy/shared-hierarchy.html)\n\n### Custom Fields from Folderless Lists\n- Table name: folderless_customfield\n- Description: Each Folderless lists can have custom fields associated with them\n- Primary key column(s):  id\n- Replicated fully or incrementally: No\n- Bookmark column(s): N/A\n- Link to API endpoint documentation: [Custom Field](https://jsapi.apiary.io/apis/clickup20/reference/0/custom-fields/get-accessible-custom-fields.html)\n\n### Custom Field from Folder Lists\n- Table name: folder_customfield\n- Description: Each Foldere list can have custom fields associated with them\n- Primary key column(s):  id\n- Replicated fully or incrementally: No\n- Bookmark column(s): N/A\n- Link to API endpoint documentation: [Custom Field](https://jsapi.apiary.io/apis/clickup20/reference/0/custom-fields/get-accessible-custom-fields.html)\n\n### Tasks\n- Table name: tasks\n- Description: Some tasks do not sit under folders. This comes from the folderless_list endpoint\n- Primary key column(s):  id\n- Replicated fully or incrementally: No\n- Bookmark column(s): date_updated. Note that the api endpoint date_updated_gt is great than or equal to, not just greater than. \n- Link to API endpoint documentation: [Get Tasks](https://jsapi.apiary.io/apis/clickup20/reference/0/tasks/get-filtered-team-tasks.html)\n\n## Other Info\n\n* Dates are returned in UNIX time \n* API Limiting uses [X-RateLimit headers](https://tools.ietf.org/id/draft-polli-ratelimit-headers-00.html)\n\n## Installation\n\n```bash\npipx install tap-clickup\n```\n\n## Usage\n\nYou can easily run `tap-clickup` by itself or in a pipeline using [Meltano](https://meltano.com/).\n\n### Executing the Tap Directly\n\n```bash\ntap-clickup --version\ntap-clickup --help\ntap-clickup --config CONFIG --discover > ./catalog.json\n```\n\n## Developer Resources\n\n### Initialize your Development Environment\n\n```bash\npipx install poetry\npoetry install\n```\n\n### Create and Run Tests\n\nCreate tests within the `tap_clickup/tests` subfolder and\n  then run:\n\n```bash\npoetry run pytest\n```\n\nYou can also test the `tap-clickup` CLI interface directly using `poetry run`:\n\n```bash\npoetry run tap-clickup --help\n```\n\n### Testing with [Meltano](https://www.meltano.com)\n\n_**Note:** This tap will work in any Singer environment and does not require Meltano.\n\nInstall Meltano (if you haven't already) and any needed plugins:\n\n```bash\n# Install meltano\npipx install meltano\n# Initialize meltano within this directory\ncd tap-clickup\nmeltano install\n```\n\nNow you can test and orchestrate using Meltano:\n\n```bash\n# Test invocation:\nmeltano invoke tap-clickup --version\n# OR run a test `elt` pipeline:\nmeltano elt tap-clickup target-jsonl\n```\n\n### SDK\n\nBuilt with the [Meltano SDK](https://sdk.meltano.com) for Singer Taps and Targets.\n\n",
    'author': 'AutoIDM',
    'author_email': None,
    'maintainer': 'AutoIDM',
    'maintainer_email': None,
    'url': 'https://autoidm.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<3.10',
}


setup(**setup_kwargs)
